from pymunk import Vec2d

import pyglet

from cocos.layer import Layer
from cocos.director import director

from roomrpg import cfg
from roomrpg.actors.enemy import Enemy


class GameObjects(Layer):
    is_event_handler = True

    def __init__(self, game, player, ui, restart_scene):
        super(GameObjects, self).__init__()

        self.game = game
        self.ui = ui
        self.player = player
        self.restart_scene = restart_scene

        self.add(self.player)

        self.enemies = {}
        self.boss = None
        self.init_enemies(game)

        self.schedule(self.loop)
        self.init_collisions(game)

        self.is_end = False

    def loop(self, dt):
        self.game.step(dt)

    def restart(self):
        director.replace(self.restart_scene(self.game))

    def init_enemies(self, game):
        for i in range(cfg.ENEMY_TYPE_0_COUNT):
            enemy = Enemy(game, index=0)
            self.enemies[enemy.cshape] = enemy
            self.add(enemy)

        for i in range(cfg.ENEMY_TYPE_1_COUNT):
            enemy = Enemy(game, index=1)
            self.enemies[enemy.cshape] = enemy
            self.add(enemy)

        self.boss = Enemy(game, index=2)
        self.enemies[self.boss.cshape] = self.boss
        self.add(self.boss)

    def get_collisions_config(self):
        return [
            (cfg.COLL_PLAYER, cfg.COLL_ENEMY, self.on_player_hit_enemy),
            (cfg.COLL_ENEMY, cfg.COLL_ENEMY, self.on_enemy_hit_enemy),
            (cfg.COLL_ENEMY, cfg.COLL_WALL, self.on_enemy_hit_wall),
            (cfg.COLL_PLAYER, cfg.COLL_WALL, self.on_player_hit_wall),
        ]

    def init_collisions(self, game):
        for coll_a, coll_b, callback in self.get_collisions_config():
            game.space.add_collision_handler(
                coll_a, coll_b, separate=callback, game=game
            )

    def clean_collision_callbacks(self):
        for coll_a, coll_b, callback in self.get_collisions_config():
            self.game.space.remove_collision_handler(coll_a, coll_b)

    def is_control(self):
        return not (self.player.is_moving or self.is_end)

    def on_key_press(self, key, modifiers):
        if not self.is_end:
            return

        if key == pyglet.window.key.SPACE:
            self.clean_collision_callbacks()
            self.game.clean_actors()
            self.restart()

    def on_mouse_motion(self, x, y, dx, dy):
        if self.is_control():
            self.player.fallow_mouse(Vec2d(x, y))

    def on_mouse_press(self, x, y, button, modifiers):
        if self.is_control():
            self.player.move_towards(Vec2d(x, y))

    def on_player_hit_enemy(self, space, arbiter, game):
        enemy = self.enemies[arbiter.shapes[1]]
        self.player.normalize_after_collision()

        status = self.player.fight(enemy)
        if status.is_kill:
            self.on_enemy_killed(game, enemy, status)
        else:
            self.on_player_killed(game, enemy, status)

    def on_enemy_hit_wall(self, space, arbiter, game):
        enemy = self.enemies[arbiter.shapes[0]]
        enemy.normalize_after_collision()

    def on_player_hit_wall(self, space, arbiter, game):
        self.player.stop_moving()

    def on_enemy_hit_enemy(self, space, arbiter, game):
        enemy0 = self.enemies[arbiter.shapes[0]]
        enemy1 = self.enemies[arbiter.shapes[1]]

        enemy0.normalize_after_collision()
        enemy1.normalize_after_collision()

    def on_enemy_killed(self, game, enemy, status):
        self.ui.update_texts()
        self.ui.show_exp_message(status.exp_delta)
        if status.lvl_delta > 0:
            self.ui.show_lvl_message(status.lvl_delta)

        enemy.kill()
        game.space.remove(enemy.cbody, enemy.cshape)

        if enemy.lvl == self.boss.lvl:
            self.is_end = True
            self.ui.show_win_message()

    def on_player_killed(self, game, enemy, status):
        self.ui.show_lost_message()
        self.is_end = True
        self.player.kill()
        game.space.remove(self.player.cbody, self.player.cshape)

        self.ui.show_killed_message(enemy.lvl)
